

# Extraction Script for Capcom Beat 'em Up Bundle

# General Extraction Process
# - Extract ARC Archive
# - Pull bin/ROMNAME file
# - Split it into parts using offsets/length
#   - Header (60b)
#   - MainCPU (???k)
#   - ??? inv gfx (???k) 
#   - AudioCPU (???k)
#   - QSound (???k)
# - Process each part
#   - maincpu: geometry
#   - gfx: geometry, deinterleave, endian swap?
#   - audiocpu: geometry
#   - qsound: geometry + endian swap

import re
import traceback
import glob
import zipfile
import logging
import os
import io

from gex.lib.archive import arc
from gex.lib.utils import blob

title = "Capcom Beat 'em Up Bundle"
description = "NYI"
in_dir_desc = "CBEUB base folder (Ex. C:\Program Files (x86)\Steam\steamapps\common\CBEUB)"

pkg_name_map = {
    "game_00.arc": "ffightj",
    "game_01.arc": "ffight",
    "game_10.arc": "kodj",
    "game_11.arc": "kod",
    "game_20.arc": "captcommj",
    "game_21.arc": "captcomm",
    "game_30.arc": "knightsj",
    "game_31.arc": "knights",
    "game_40.arc": "wofj",
    "game_41.arc": "wof",
    "game_50.arc": "pgear",
    "game_51.arc": "armwar",
    "game_60.arc": "batcirj",
    "game_61.arc": "batcir",
}

def write_temp_file(contents, path):
    with open(path, 'wb') as out_file:
        out_file.write(contents)

def find_files(base_path):
    arc_path = os.path.join(base_path, "nativeDX11x64", "arc") 
    candidate_files = glob.glob(arc_path +'/game_*.arc')
    archive_list = []
    for candidate in candidate_files:
        if re.search(r'game_\d\d.arc', candidate):
            archive_list.append(candidate)
    return archive_list

def placeholder_generator(file_map):
    def create_placeholders(contents):
        out_files = {}
        for filename, size in file_map.items():
            out_files[filename] = bytes(size*b'\0')
        return out_files  
    return create_placeholders


def deshuffle_gfx_common(start, length, filenames, num_deinterleave_split, do_split):
    def gfx(contents):
        # Cut out the section
        contents = contents[start:start+length]

        # This is weird... it's a bit shuffle, not byte-level and not a normal interleave
        bit_order = [
            7, 3, 15, 11, 23, 19, 31, 27,
            6, 2, 14, 10, 22, 18, 30, 26,
            5, 1, 13, 9, 21, 17, 29, 25,
            4, 0, 12, 8, 20, 16, 28, 24,
            39, 35, 47, 43, 55, 51, 63, 59,
            38, 34, 46, 42, 54, 50, 62, 58, 
            37, 33, 45, 41, 53, 49, 61, 57, 
            36, 32, 44, 40, 52, 48, 60, 56
        ]
        chunks = blob.split_bit_shuffle(contents, word_size_bytes=8, bit_order=bit_order, num_ways=num_deinterleave_split)

        # Split it
        if do_split:
            new_chunks = []
            for oldchunk in chunks:
                new_chunks.extend(blob.equal_split(oldchunk, num_chunks = 2))
            chunks = new_chunks

        return dict(zip(filenames, chunks))
    return gfx

    
def audio_common(start, filenames):
    def audio(contents):
        chunks = []

        # Add the audio CPU
        chunks.append(contents[start:start+0x8000] + contents[start+0x10000:start+0x18000])

        # Add the qsound
        qsound_start = start+0x18000
        qsound_contents = contents[qsound_start:qsound_start+0x40000]
        chunks.extend(blob.equal_split(qsound_contents, num_chunks=2))

        return dict(zip(filenames, chunks))
    return audio


################################################################################
# START Final Fight                                                            #
################################################################################
# game_00.arc: Final Fight (JP)
# game_01.arc: Final Fight

def handle_ffight(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [
        "ff_42.11h",
        "ffe_43.12h",
        "ff_36.11f",
        "ff_37.12f",
        "ff-32m.8h"
    ]
    def maincpu(contents):
        chunk_5 = contents[0x080040:0x100040]
        contents = contents[0x40:0x080040]
        chunks = blob.deinterleave(contents, num_ways=2, word_size=1)
        
        new_chunks = []
        for oldchunk in chunks:
            new_chunks.extend(blob.equal_split(oldchunk, num_chunks = 2))
        chunks = new_chunks

        # Add 5th non-interleaved chunk
        chunks.append(chunk_5)

        return dict(zip(maincpu_filenames, chunks))
    func_map['maincpu'] = maincpu

    gfx_filenames = [
        "ff-5m.7a",
        "ff-7m.9a",
        "ff-1m.3a",
        "ff-3m.5a"
    ]
    func_map['gfx'] = deshuffle_gfx_common(0x400040, 0x200000, gfx_filenames, 4, False)

    audio_filenames = [
        'ff_09.12b',
        'ff_18.11c',
        'ff_19.12c'
    ]
    func_map['audio'] = audio_common(0x600040, audio_filenames)

    ph_files = {
        'buf1': 0x117,
        'ioa1': 0x117,
        'prg1': 0x117,
        'rom1': 0x117,
        'sou1': 0x117,
        's224b.1a': 0x117,
        'iob1.11e': 0x117
        }
    func_map['placeholders'] = placeholder_generator(ph_files)

    zip_contents = merged_rom_handler(merged_contents, func_map)
    out_files.append({'filename': 'ffight.zip', 'contents': zip_contents})

    return out_files


def handle_ffightj(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [
        "ff42.bin",
        "ff43.bin",
        "ffj_40.10h",
        "ffj_41.11h",
        "ff36.bin",
        "ff37.bin",
        "ffj_34.10f",
        "ffj_35.11f"
    ]
    def maincpu(contents):
        contents = contents[0x40:0x100040]
        chunks = blob.deinterleave(contents, num_ways=2, word_size=1)
        
        new_chunks = []
        for oldchunk in chunks:
            new_chunks.extend(blob.equal_split(oldchunk, num_chunks = 4))
        chunks = new_chunks
        return dict(zip(maincpu_filenames, chunks))
    func_map['maincpu'] = maincpu

    gfx_filenames = [
        "ffj_09.4b",
        "ffj_10.5b",
        "ffj_01.4a",
        "ffj_02.5a",
        "ffj_13.9b",
        "ffj_14.10b",
        "ffj_05.9a",
        "ffj_06.10a",
        "ffj_24.5e",
        "ffj_25.7e",
        "ffj_17.5c",
        "ffj_18.7c",
        "ffj_38.8h",
        "ffj_39.9h",
        "ffj_32.8f",
        "ffj_33.9f"
    ]
    func_map['gfx'] = deshuffle_gfx_common(0x400040, 0x200000, gfx_filenames, 8, True)

    audio_filenames = [
        'ff_23.bin',
        'ffj_30.bin',
        'ffj_31.bin'
    ]
    func_map['audio'] = audio_common(0x600040, audio_filenames)
    
    ph_files = {
        'buf1': 0x117,
        'ioa1': 0x117,
        'prg1': 0x117,
        'rom1': 0x117,
        'sou1': 0x117,
        's222b.1a': 0x117,
        'lwio.12c': 0x117
        }
    func_map['placeholders'] = placeholder_generator(ph_files)
    
    zip_contents = merged_rom_handler(merged_contents, func_map)
    out_files.append({'filename': 'ffightj.zip', 'contents': zip_contents})

    return out_files

################################################################################
# END Final Fight                                                              #
################################################################################


################################################################################
# START The King of Dragons                                                    #
################################################################################
# game_10.arc: The King of Dragons (JP)
# game_11.arc: The King of Dragons


def handle_kod(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [
        "kde_37a.11f", 
        "kde_38a.12f", 
        "kd_35.9f", 
        "kd_36a.10f",
        "kde_30a.11e", 
        "kde_31a.12e", 
        "kd_28.9e", 
        "kd_29.10e"
    ]
    def maincpu(contents):
        contents = contents[0x40:0x100040]
        chunks = blob.deinterleave(contents, num_ways=2, word_size=1)
        
        new_chunks = []
        for oldchunk in chunks:
            new_chunks.extend(blob.equal_split(oldchunk, num_chunks = 4))
        chunks = new_chunks

        return dict(zip(maincpu_filenames, chunks))
    func_map['maincpu'] = maincpu

    gfx_filenames = [
        "kd-5m.4a",
        "kd-6m.4c",
        "kd-7m.6a",
        "kd-8m.6c",
        "kd-1m.3a",
        "kd-2m.3c",
        "kd-3m.5a",
        "kd-4m.5c"
    ]
    func_map['gfx'] = deshuffle_gfx_common(0x400040, 0x400000, gfx_filenames, 4, True)

    audio_filenames = [
        'kd_9.12a',
        'kd_18.11c',
        'kd_19.12c'
    ]
    func_map['audio'] = audio_common(0x800040, audio_filenames)

    ph_files = {
        'buf1': 0x117,
        'ioa1': 0x117,
        'prg1': 0x117,
        'rom1': 0x117,
        'sou1': 0x117,
        'kd29b.1a': 0x117,
        'iob1.11d': 0x117,
        'ioc1.ic7': 0x104,
        'c632.ic1': 0x117
    }
    func_map['placeholders'] = placeholder_generator(ph_files)

    zip_contents = merged_rom_handler(merged_contents, func_map)
    out_files.append({'filename': 'kod.zip', 'contents': zip_contents})

    return out_files



def handle_kodj(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [
        "kdj_37a.11f",
        "kdj_38a.12f",
        "kdj_30a.11e",
        "kdj_31a.12e",
        "kd_33.6f"
    ]
    def maincpu(contents):
        chunk_5 = contents[0x080040:0x100040]
        contents = contents[0x40:0x080040]
        chunks = blob.deinterleave(contents, num_ways=2, word_size=1)
        
        new_chunks = []
        for oldchunk in chunks:
            new_chunks.extend(blob.equal_split(oldchunk, num_chunks = 2))
        chunks = new_chunks

        # Add 5th non-interleaved chunk
        chunks.append(chunk_5)

        return dict(zip(maincpu_filenames, chunks))
    func_map['maincpu'] = maincpu

    gfx_filenames = [
        "kd_06.8a",
        "kd_15.8c",
        "kd_08.10a",
        "kd_17.10c",
        "kd_05.7a",
        "kd_14.7c",
        "kd_07.9a",
        "kd_16.9c"
    ]
    func_map['gfx'] = deshuffle_gfx_common(0x400040, 0x400000, gfx_filenames, 4, True)

    audio_filenames = [
        'kd_09.12a',
        'kd_18.11c',
        'kd_19.12c'
    ]
    func_map['audio'] = audio_common(0x800040, audio_filenames)

    ph_files = {
        'buf1': 0x117,
        'ioa1': 0x117,
        'prg1': 0x117,
        'rom1': 0x117,
        'sou1': 0x117,
        'kd29b.1a': 0x117,
        'iob1.11d': 0x117,
        'ioc1.ic7': 0x104,
        'c632.ic1': 0x117
    }
    func_map['placeholders'] = placeholder_generator(ph_files)

    zip_contents = merged_rom_handler(merged_contents, func_map)
    out_files.append({'filename': 'kodj.zip', 'contents': zip_contents})

    return out_files


################################################################################
# END The King of Dragons                                                      #
################################################################################


################################################################################
# START Captain Commando                                                       #
################################################################################
# game_20.arc: Captain Commando (JP)
# game_21.arc: Captain Commando

def handle_captcomm(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [
        "cce_23f.8f",
        "cc_28f.9f",
        "cc_22f.7f",
        "cc_24f.9e"
    ]
    def maincpu(contents):
        # Only the last 2 128k chunks actually need deinterleaved...
        maincpu_area = contents[0x40:0x140040]
        deint_chunks = blob.deinterleave(maincpu_area[0x100000:0x140000], num_ways=2, word_size=1)

        chunks = []
        chunks.append(maincpu_area[0x0:0x80000])
        chunks.append(deint_chunks[0])
        chunks.append(maincpu_area[0x80000:0x100000])
        chunks.append(deint_chunks[1])

        return dict(zip(maincpu_filenames, chunks))
    func_map['maincpu'] = maincpu

    gfx_filenames = [
        "cc-5m.3a",
        "cc-6m.7a",
        "cc-7m.5a",
        "cc-8m.9a",
        "cc-1m.4a",
        "cc-2m.8a",
        "cc-3m.6a",
        "cc-4m.10a"
    ]
    func_map['gfx'] = deshuffle_gfx_common(0x400040, 0x400000, gfx_filenames, 4, True)

    audio_filenames = [
        'cc_09.11a',
        'cc_18.11c',
        'cc_19.12c'
    ]
    func_map['audio'] = audio_common(0x800040, audio_filenames)

    ph_files = {
        'buf1': 0x117,
        'ioa1': 0x117,
        'prg1': 0x117,
        'rom1': 0x117,
        'sou1': 0x117,
        'cc63b.1a': 0x117,
        'iob1.12d': 0x117,
        'ccprg1.11d': 0x117,
        'ioc1.ic7': 0x104,
        'c632b.ic1': 0x117
    }
    func_map['placeholders'] = placeholder_generator(ph_files)

    zip_contents = merged_rom_handler(merged_contents, func_map)
    out_files.append({'filename': 'captcomm.zip', 'contents': zip_contents})

    return out_files



def handle_captcommj(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [
        "ccj_23f.8f",
        "ccj_28f.9f",
        "ccj_22f.7f",
        "ccj_24f.9e"
    ]
    def maincpu(contents):
        # Only the last 2 128k chunks actually need deinterleaved...
        maincpu_area = contents[0x40:0x140040]
        deint_chunks = blob.deinterleave(maincpu_area[0x100000:0x140000], num_ways=2, word_size=1)

        chunks = []
        chunks.append(maincpu_area[0x0:0x80000])
        chunks.append(deint_chunks[0])
        chunks.append(maincpu_area[0x80000:0x100000])
        chunks.append(deint_chunks[1])

        return dict(zip(maincpu_filenames, chunks))
    func_map['maincpu'] = maincpu

    gfx_filenames = [
        "cc_01.3a",
        "cc_05.7a",
        "cc_02.4a",
        "cc_06.8a",
        "cc_03.5a",
        "cc_07.9a",
        "cc_04.6a",
        "cc_08.10a"
    ]
    func_map['gfx'] = deshuffle_gfx_common(0x400040, 0x400000, gfx_filenames, 4, True)
    
    audio_filenames = [
        'ccj_09.12a',
        'ccj_18.11c',
        'ccj_19.12c'
    ]
    func_map['audio'] = audio_common(0x800040, audio_filenames)

    ph_files = {
        'buf1': 0x117,
        'ioa1': 0x117,
        'prg1': 0x117,
        'rom1': 0x117,
        'sou1': 0x117,
        'cc63b.1a': 0x117,
        'iob1.12d': 0x117,
        'ccprg1.11d': 0x117,
        'ioc1.ic7': 0x104,
        'c632.ic1': 0x117
    }
    func_map['placeholders'] = placeholder_generator(ph_files)

    zip_contents = merged_rom_handler(merged_contents, func_map)
    out_files.append({'filename': 'captcommj.zip', 'contents': zip_contents})

    return out_files


################################################################################
# END Captain Commando                                                         #
################################################################################


################################################################################
# START Knights of the Round                                                   #
################################################################################
# game_30.arc: Knights of the Round (JP)
# game_31.arc: Knights of the Round



def handle_knights(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [
        "kr_23e.8f",
        "kr_22.7f"
    ]
    def maincpu(contents):
        # Only the last 2 128k chunks actually need deinterleaved...
        maincpu_area = contents[0x40:0x100040]
        chunks = blob.equal_split(maincpu_area, num_chunks=2)

        return dict(zip(maincpu_filenames, chunks))
    func_map['maincpu'] = maincpu

    gfx_filenames = [
        "kr-5m.3a",
        "kr-6m.7a",
        "kr-7m.5a",
        "kr-8m.9a",
        "kr-1m.4a",
        "kr-2m.8a",
        "kr-3m.6a",
        "kr-4m.10a"
    ]
    func_map['gfx'] = deshuffle_gfx_common(0x400040, 0x400000, gfx_filenames, 4, True)

    audio_filenames = [
        'kr_09.11a',
        'kr_18.11c',
        'kr_19.12c'
    ]
    func_map['audio'] = audio_common(0x800040, audio_filenames)

    ph_files = {
        'buf1': 0x117,
        'ioa1': 0x117,
        'prg1': 0x117,
        'rom1': 0x117,
        'sou1': 0x117,
        'kr63b.1a': 0x117,
        'iob1.12d': 0x117,
        'bprg1.11d': 0x117,
        'ioc1.ic7': 0x104,
        'c632.ic1': 0x117
    }
    func_map['placeholders'] = placeholder_generator(ph_files)

    zip_contents = merged_rom_handler(merged_contents, func_map)
    out_files.append({'filename': 'knights.zip', 'contents': zip_contents})

    return out_files



def handle_knightsj(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [
        "kr_23j.8f",
        "kr_22.7f"
    ]
    def maincpu(contents):
        # Only the last 2 128k chunks actually need deinterleaved...
        maincpu_area = contents[0x40:0x100040]
        chunks = blob.equal_split(maincpu_area, num_chunks=2)

        return dict(zip(maincpu_filenames, chunks))
    func_map['maincpu'] = maincpu

    gfx_filenames = [
        "kr_01.3a",
        "kr_05.7a",
        "kr_02.4a",
        "kr_06.8a",
        "kr_03.5a",
        "kr_07.9a",
        "kr_04.6a",
        "kr_08.10a"
    ]
    func_map['gfx'] = deshuffle_gfx_common(0x400040, 0x400000, gfx_filenames, 4, True)

    audio_filenames = [
        'kr_09.12a',
        'kr_18.11c',
        'kr_19.12c'
    ]
    func_map['audio'] = audio_common(0x800040, audio_filenames)

    ph_files = {
        'buf1': 0x117,
        'ioa1': 0x117,
        'prg1': 0x117,
        'rom1': 0x117,
        'sou1': 0x117,
        'kr63b.1a': 0x117,
        'iob1.12d': 0x117,
        'bprg1.11d': 0x117,
        'ioc1.ic7': 0x104,
        'c632.ic1': 0x117
    }
    func_map['placeholders'] = placeholder_generator(ph_files)

    zip_contents = merged_rom_handler(merged_contents, func_map)
    out_files.append({'filename': 'knightsj.zip', 'contents': zip_contents})

    return out_files


################################################################################
# END Knights of the Round                                                     #
################################################################################


################################################################################
# START Warriors of Fate                                                       #
################################################################################
# game_40.arc: Warriors of Fate (JP)
# game_41.arc: Warriors of Fate

# wof map from script
#           0x000000    0x000040                IBIS Header
# maincpu   0x000040    0x100040                OK
#           0x100040    0x400040                Padding (All FF)
# gfx       0x400040    0x800040                OK
# audiocpu  0x800040    0x828040   (with gap)   BAD CRC, but looks like the audiocpu code from other ROMS (version header)
#           |
#           |  0x800040    0x808040                Different data - does not look like most audiocpu - no version header, a lot of it looks like junk filler
#           |  0x808040    0x810040                Padding (All FF) - similar padding is only 0x075DB - 0x08000 in tk2_qa.5k
#           |  0x810040    0x828040                Matches 0x08000 - 0x20000 in wof's tk2_qa.5k
#           0x828040    0x830040                Unknown
#           0x830040    0x850040                Padding (All FF)
# qsound    0x850040    0xA50040                OK


def wof_audio_common(filenames):
    def audio(contents):
        start = 0x800040
        chunks = []

        # Add the audio CPU
        audiocpu_content = contents[start:start+0x8000] + contents[start+0x10000:start+0x28000]
        chunks.append(audiocpu_content)

        # Add the qsound
        qsound_start = start+0x50000
        qsound_contents = contents[qsound_start:qsound_start+0x200000]
        chunks.extend(blob.equal_split(qsound_contents, num_chunks=4))

        return dict(zip(filenames, chunks))
    return audio

def handle_wof(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [
        "tk2e_23c.8f",
        "tk2e_22c.7f"
    ]
    def maincpu(contents):
        maincpu_area = contents[0x40:0x100040]
        chunks = blob.equal_split(maincpu_area, num_chunks=2)

        return dict(zip(maincpu_filenames, chunks))
    func_map['maincpu'] = maincpu

    gfx_filenames = [
        "tk2-1m.3a",
        "tk2-5m.7a",
        "tk2-3m.5a",
        "tk2-7m.9a",
        "tk2-2m.4a",
        "tk2-6m.8a",
        "tk2-4m.6a",
        "tk2-8m.10a"
    ]

    func_map['gfx'] = deshuffle_gfx_common(0x400040, 0x400000, gfx_filenames, 4, True)

    audio_filenames = [
        'tk2_qa.5k',
        'tk2-q1.1k',
        'tk2-q2.2k',
        'tk2-q3.3k',
        'tk2-q4.4k'
    ]
    func_map['audio'] = wof_audio_common(audio_filenames)

    ph_files = {
        'buf1': 0x117,
        'ioa1': 0x117,
        'prg2': 0x117,
        'rom1': 0x117,
        'tk263b.1a': 0x117,
        'iob1.12d': 0x117,
        'bprg1.11d': 0x117,
        'ioc1.ic1': 0x104,
        'd7l1.7l': 0x117,
        'd8l1.8l': 0x117,
        'd9k1.9k': 0x117,
        'd10f1.10f': 0x117
    }
    func_map['placeholders'] = placeholder_generator(ph_files)

    zip_contents = merged_rom_handler(merged_contents, func_map)
    out_files.append({'filename': 'wof.zip', 'contents': zip_contents})

    return out_files

def handle_wofj(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [
        "tk2j_23c.8f",
        "tk2j_22c.7f"
    ]
    def maincpu(contents):
        maincpu_area = contents[0x40:0x100040]
        chunks = blob.equal_split(maincpu_area, num_chunks=2)

        return dict(zip(maincpu_filenames, chunks))
    func_map['maincpu'] = maincpu

    gfx_filenames = [
        "tk2_01.3a",
        "tk2_05.7a",
        "tk2_02.4a",
        "tk2_06.8a",
        "tk2_03.5a",
        "tk2_07.9a",
        "tk2_04.6a",
        "tk2_08.10a"
    ]

    func_map['gfx'] = deshuffle_gfx_common(0x400040, 0x400000, gfx_filenames, 4, True)

    audio_filenames = [
        'tk2_qa.5k',
        'tk2-q1.1k',
        'tk2-q2.2k',
        'tk2-q3.3k',
        'tk2-q4.4k'
    ]
    func_map['audio'] = wof_audio_common(audio_filenames)

    ph_files = {
        'buf1': 0x117,
        'ioa1': 0x117,
        'prg1': 0x117,
        'sou1': 0x117,
        'rom1': 0x117,
        'tk263b.1a': 0x117,
        'iob1.12d': 0x117,
        'bprg1.11d': 0x117,
        'ioc1.ic1': 0x104,
        'd7l1.7l': 0x117,
        'd8l1.8l': 0x117,
        'd9k1.9k': 0x117,
        'd10f1.10f': 0x117
    }
    func_map['placeholders'] = placeholder_generator(ph_files)

    zip_contents = merged_rom_handler(merged_contents, func_map)
    out_files.append({'filename': 'wofj.zip', 'contents': zip_contents})

    return out_files

################################################################################
# END Warriors of Fate                                                         #
################################################################################


################################################################################
# START Armored Warriors                                                       #
################################################################################
# game_50.arc: Powered Gear: Strategic Variant Armor Equipment (JP)
# game_51.arc: Armored Warriors

def armwar_gfx_common(contents):
    # Cut out the section
    contents = contents[0x0800040:0x1C00040]

    # This is weird... it's a bit shuffle, not byte-level and not a normal interleave
    bit_order = [7, 3, 15, 11, 23, 19, 31, 27, 6, 2, 14, 10, 22, 18, 30, 26, 5, 1, 13, 9, 21, 17, 29, 25, 4, 0, 12, 8, 20, 16, 28, 24]
    contents = blob.bit_shuffle(contents, word_size_bytes=4, bit_order=bit_order)

    # Split it
    chunks = blob.equal_split(contents, num_chunks=20)

    # Interleave each pair of chunks
    new_chunks = []
    for oddchunk,evenchunk in zip(chunks[0::2], chunks[1::2]):
        new_chunks.append(blob.interleave([oddchunk, evenchunk], word_size=8))
    chunks = new_chunks

    # Merge the chunks back together
    contents = blob.merge(chunks)

    # Deinterleave the chunks into our files
    new_chunks = []
    chunks = blob.deinterleave(contents, num_ways = 4, word_size=2)
    for chunk in chunks:
        new_chunks.extend(blob.custom_split(chunk, [0x400000, 0x100000]))
    chunks = new_chunks
    filenames = [
        'pwg.13m',
        'pwg.14m',
        'pwg.15m',
        'pwg.16m',
        'pwg.17m',
        'pwg.18m',
        'pwg.19m',
        'pwg.20m'
    ]
    return dict(zip(filenames, chunks))

def armwar_audiocpu_common(contents):
    chunks = []
    chunks.append(contents[0x1C00040:0x1C08040] + contents[0x1C10040:0x1C28040])
    chunks.append(contents[0x1C28040:0x1C48040])
    filenames = [
        'pwg.01',
        'pwg.02'
    ]
    return dict(zip(filenames, chunks))
    
def armwar_qsound_common(contents):
    contents = contents[0x1C50040:0x2050040]
    chunks = blob.equal_split(contents, num_chunks=2)
    chunks = blob.swap_endian_all(chunks)
    filenames = [
        'pwg.11m',
        'pwg.12m'
    ]
    return dict(zip(filenames, chunks))

def handle_armwar(merged_contents):
    out_files = []

    def maincpu(contents):
        contents = contents[0x40:0x400040]
        chunks = blob.equal_split(contents, num_chunks=8)
        filenames = [   
            "pwge.03c",
            "pwge.04c",
            "pwge.05b",
            "pwg.06",
            "pwg.07",
            "pwg.08",
            "pwg.09a",
            "pwg.10"
        ]
        return dict(zip(filenames, chunks))
    func_map = {}
    func_map['maincpu'] = maincpu
    func_map['gfx'] = armwar_gfx_common
    func_map['audiocpu'] = armwar_audiocpu_common
    func_map['qsound'] = armwar_qsound_common
    out_files.append({'filename': 'armwar.zip', 'contents': merged_rom_handler(merged_contents, func_map)})
    return out_files

def handle_pgear(merged_contents):
    out_files = []

    def maincpu(contents):
        contents = contents[0x40:0x400040]
        chunks = blob.equal_split(contents, num_chunks=8)
        filenames = [   
            "pwgj.03a",
            "pwgj.04a",
            "pwgj.05a",
            "pwg.06",
            "pwg.07",
            "pwg.08",
            "pwg.09a",
            "pwg.10"
        ]
        return dict(zip(filenames, chunks))
    func_map = {}
    func_map['maincpu'] = maincpu
    func_map['gfx'] = armwar_gfx_common
    func_map['audiocpu'] = armwar_audiocpu_common
    func_map['qsound'] = armwar_qsound_common
    out_files.append({'filename': 'pgear.zip', 'contents': merged_rom_handler(merged_contents, func_map)})
    return out_files

################################################################################
# END Armored Warriors                                                         #
################################################################################

################################################################################
# START Battle Circuit                                                         #
################################################################################
# game_60.arc: Battle Circuit (JP)
# game_61.arc: Battle Circuit
# Battle Circuit has some weird graphics interleaving that took some time to recreate from the Japanese script.
# None the less, it's in - and only the maincpu differs between the two!

def batcir_gfx_common(contents):
    # Cut out the section
    contents = contents[0x0800040:0x1800040]

    # This is weird... it's a bit shuffle, not byte-level and not a normal interleave
    bit_order = [7, 3, 15, 11, 23, 19, 31, 27, 6, 2, 14, 10, 22, 18, 30, 26, 5, 1, 13, 9, 21, 17, 29, 25, 4, 0, 12, 8, 20, 16, 28, 24]
    contents = blob.bit_shuffle(contents, word_size_bytes=4, bit_order=bit_order)

    # Split it
    chunks = blob.equal_split(contents, num_chunks=16)

    # Interleave each pair of chunks
    new_chunks = []
    for oddchunk,evenchunk in zip(chunks[0::2], chunks[1::2]):
        new_chunks.append(blob.interleave([oddchunk, evenchunk], word_size=8))
    chunks = new_chunks

    # Merge the chunks back together
    contents = blob.merge(chunks)

    # Deinterleave the chunks into our 4 files
    chunks = blob.deinterleave(contents, num_ways = 4, word_size=2)
    filenames = [
        'btc.13m',
        'btc.15m',
        'btc.17m',
        'btc.19m'
    ]
    return dict(zip(filenames, chunks))

def batcir_audiocpu_common(contents):
    chunks = []
    chunks.append(contents[0x1800040:0x1808040] + contents[0x1810040:0x1828040])
    chunks.append(contents[0x1828040:0x1848040])
    filenames = [
        'btc.01',
        'btc.02'
    ]
    return dict(zip(filenames, chunks))
    
def batcir_qsound_common(contents):
    contents = contents[0x1850040:0x1C50040]
    chunks = blob.equal_split(contents, num_chunks=2)
    chunks = blob.swap_endian_all(chunks)
    filenames = [
        'btc.11m',
        'btc.12m'
    ]
    return dict(zip(filenames, chunks))


def handle_batcir(merged_contents):
    out_files = []

    def maincpu(contents):
        contents = contents[0x40:0x380040]
        chunks = blob.equal_split(contents, num_chunks=7)
        filenames = [   
            "btce.03", 
            "btce.04", 
            "btce.05", 
            "btce.06", 
            "btc.07", 
            "btc.08", 
            "btc.09"
        ]
        return dict(zip(filenames, chunks))
    func_map = {}
    func_map['maincpu'] = maincpu
    func_map['gfx'] = batcir_gfx_common
    func_map['audiocpu'] = batcir_audiocpu_common
    func_map['qsound'] = batcir_qsound_common
    out_files.append({'filename': 'batcir.zip', 'contents': merged_rom_handler(merged_contents, func_map)})
    return out_files


def handle_batcirj(merged_contents):
    out_files = []

    def maincpu(contents):
        contents = contents[0x40:0x380040]
        chunks = blob.equal_split(contents, num_chunks=7)
        filenames = [   
            "btcj.03", 
            "btcj.04", 
            "btcj.05", 
            "btcj.06", 
            "btc.07", 
            "btc.08", 
            "btc.09"
        ]
        return dict(zip(filenames, chunks))
    func_map = {}
    func_map['maincpu'] = maincpu
    func_map['gfx'] = batcir_gfx_common
    func_map['audiocpu'] = batcir_audiocpu_common
    func_map['qsound'] = batcir_qsound_common
    out_files.append({'filename': 'batcirj.zip', 'contents': merged_rom_handler(merged_contents, func_map)})
    return out_files


################################################################################
# END Battle Circuit                                                           #
################################################################################


def merged_rom_handler(merged_contents, func_map):
    new_data = dict()
    for func in func_map.values():
        new_data.update(func(bytearray(merged_contents)))

    # Build the new zip file
    new_contents = io.BytesIO()
    with zipfile.ZipFile(new_contents, "w") as new_archive:
        for name, data in new_data.items():
            new_archive.writestr(name, data)
    return new_contents.getvalue()


def main(game_base_dir, out_path):
    pak_files = find_files(game_base_dir)
    for file_path in pak_files:
        file_name = os.path.basename(file_path)
        pkg_name = pkg_name_map[file_name]
        logging.info(f"Extracting {file_name}: {pkg_name}") 
        try:
            with open(file_path, "rb") as curr_file:
                file_content = bytearray(curr_file.read())
                arc_contents = arc.extract(file_content)
                output_files = []

                # Get the bin entry
                merged_rom_contents = None
                for key, arc_content in arc_contents.items():
                    if arc_content['path'].startswith('bin'):
                        merged_rom_contents = arc_content['contents']

                handler_func = globals().get(f'handle_{pkg_name}')

                if merged_rom_contents != None and handler_func != None:
                    output_files = handler_func(merged_rom_contents)
                        
                    for output_file in output_files:
                        with open(os.path.join(out_path, output_file['filename']), "wb") as out_file:
                            out_file.write(output_file['contents'])
                elif merged_rom_contents == None:
                    print("Could not find merged rom data in arc.")
                elif handler_func == None:
                    print("Could not find matching handler function.")
        except Exception as e:
            traceback.print_exc()
            logging.warning(f'Error while processing {file_path}!') 

    logging.info("""
        Processing complete. 
    """)



import click
import sys
import zlib

@click.command()
@click.option('--in', 'in_file', help = 'path to input file', required=True)
@click.option('--out', 'out_file', help = 'path to output file', required=True)
@click.option('--start', help = 'position to start in file', required=True, default=0, type=int)
@click.option('--length', help = 'length to copy in bytes', required=True, type=int)
def slice(in_file, out_file, start, length):
    """Slice a portion of a file to a new file"""

    def read_bin_file(path):
        try: 
            with open(path, "rb") as f:
                content = f.read()
                return content
        except IOError:
            print(f"Error reading {path}!")   

    def write_bin_file(data, path):
        try: 
            with open(path, "wb") as f:
                f.write(data)
        except IOError:
            print(f"Error writing {path}!")  


    print(f"Pulling {length} bytes from {in_file} starting at {start}")
    print(f"Saving to {out_file}")

    in_data = read_bin_file(in_file)
    out_data = in_data[int(start):int(start + length)]
    write_bin_file(out_data, out_file)
    crc_val = zlib.crc32(out_data)

    print(f"Saved to {out_file} and calculated CRC {hex(crc_val)}")


import click
import os


@click.command()
@click.option('--target', 'target_dir', help = 'path to directory of files to search for', required=True)
@click.option('--search', 'search_dir', help = 'path to directory of files to search within', required=True)
def bulk(target_dir, search_dir):
    """Find target files with offsets within the files of a search path"""

    file_cache = {}

    def precache_file(file_path):
        file_data = read_bin_file(file_path)
        file_cache[file_path] = file_data

    def read_bin_file(path):
        try: 
            with open(path, "rb") as f:
                content = f.read()
                return content
        except IOError:
            print('Error While Opening the file!')  
    
    def find_match(archive_content, sub_content):
        for i in range(0x0, len(archive_content) - len(sub_content)):
            if archive_content[i] == sub_content[0]:
                valid = True
                for j in range(0, len(sub_content)-1):
                    if archive_content[i+j] != sub_content[j]:
                        valid = False
                        break
                if valid:
                    return {'start': i, 'end': i+len(sub_content)}
        return None
        
    def list_child_files(child_path):
        child_files = []
        for (dirpath, dirnames, filenames) in os.walk(child_path):
            for filename in filenames:
                child_files.append(os.path.join(child_path, filename))
            break
        return child_files

    search_files = list_child_files(search_dir)
    target_files = list_child_files(target_dir)

    matches = []
        
    print(f"Precaching...")
    for target_file in target_files:
        print(f" Precaching {target_file}...")
        precache_file(target_file)

    print(f"After caching:")

    print(f"Searching...")
    for search_file in search_files:
        print(f" Searching {search_file}...")
        search_content = read_bin_file(search_file)

        for target_file in target_files:
            target_content = file_cache[target_file]
            match = find_match(search_content, target_content)
            if match == None:
                print(f"  No match found for {target_file}") 
            else:
                print(f"  Match for {target_file} from {hex(match['start'])} to {hex(match['end'])}!")
                matches.append(match)
                


import click
import os
import glob
import importlib
import pkgutil
from gex.lib.transforms import *
import inspect

@click.command()
def list():
    """List Available Tasks"""
    for global_name, transform_module in globals().items():
        if inspect.ismodule(transform_module) and transform_module.__package__ == 'gex.lib.transforms':
            print(f'{global_name}: {transform_module.title}')
            print(f'  {transform_module.description}')
            print(f'  Expected input dir: {transform_module.in_dir_desc}')



import click

from .steam.steam import steam
from .wii.wii import wii

@click.group()
def postprocess():
    """Cleanup Processing for Various Rip Types"""
    pass

postprocess.add_command(steam)
postprocess.add_command(wii)


import click
import os
import re
import zlib


def list_files(dir):
    files = []
    for (dirpath, dirnames, filenames) in os.walk(dir):
        for filename in filenames:
            files.append(os.path.join(dir, filename))
        break
    return files

@click.command()
@click.option('--srcdir', 'src_dir', help = 'path to directory with cleanrip output files', required=True)
@click.option('--destdir', 'dest_dir', help = 'path to send reassembled, verified ROMs to', required=True)
def cleanrip(src_dir, dest_dir):
    """Clean up Cleanrip dumps"""

    print("NYI")
    # List files in output dir
    src_files = list_files(src_dir)

    # Build output sets
    sets = dict()
    for src_file in src_files:
        print(f"Handling {src_file}...")
        filename = os.path.basename(src_file)
        
        # Get the disc ID and find/make the set object
        disc_id_match = re.match("([A-Z0-9]{6})", filename)
        if disc_id_match != None:
            disc_id = disc_id_match.group(0)
            current_set = None
            if disc_id in sets:
                current_set = sets[disc_id]
            else:
                current_set = dict()
                sets[disc_id] = current_set


            if filename.endswith(".bca"):
                current_set["bca"] = filename
            elif filename.endswith("-dumpinfo.txt"):
                current_set["dumpinfo"] = filename
            else:
                part_match = re.search("(?<=part)(\d)(?=.iso\Z)", filename)
                if part_match != None:
                    part_array = None
                    if "parts" in current_set:
                        part_array = current_set["parts"]
                    else:
                        part_array = []
                        current_set["parts"] = part_array
                    part_array.append(filename)
        else:
            print(f"Skipping {filename} since it doesn't match name pattern...")

    # Verify valid sets
    invalid_set_keys = []
    for set_id in sets.keys():
        current_set = sets[set_id]
        if not "dumpinfo" in current_set:
            # We need a dumpinfo to get data
            print(f"No dumpinfo for set {set_id} - skipping set!")
            invalid_set_keys.append(set_id)
        else:
            # Check the part array for continuity
            part_array = current_set["parts"]
            part_array.sort()
            for part_idx, part_filename in enumerate(part_array):
                if part_filename != f"{set_id}.part{part_idx}.iso":
                    print(f"{set_id}: {part_idx} found {part_filename}, which is incorrct; skipping set!")
                    invalid_set_keys.append(set_id)
                    break
            
    for invalid_set_key in invalid_set_keys:
        del sets[invalid_set_key]

    # Get game name and hash from XXXXXX-dumpinfo.txt
    for set_id in sets.keys():
        current_set = sets[set_id]
        print(current_set)
        with open(os.path.join(src_dir, current_set["dumpinfo"])) as file:
            lines = file.readlines()
            lines = [line.rstrip() for line in lines]
            print(lines)
            for line in lines:
                if line.startswith("Internal Name: "):
                    current_set["name"] = line[len("Internal Name: "):]
                    safename = current_set["name"]
                    safename = re.sub(r' +', '_', safename)
                    safename = re.sub(r'[\W]+', '', safename)
                    current_set["safename"] = safename
                elif line.startswith("CRC32: "):
                    current_set["crc"] = line[len("CRC32: "):]

    print("Found sets: ")
    for set_id in sets.keys():
        current_set = sets[set_id]
        print(f"  {set_id}")
        print(f"    {current_set['name']}")
        print(f"    {current_set['safename']}")
        print(f"    {current_set['parts']}")
        print(f"    {current_set['crc']}")
    
    for set_id in sets.keys():
        current_set = sets[set_id]
        print(f"Reading parts for {current_set['name']}...")
        content = bytes()
        for partname in current_set['parts']:
            with open(os.path.join(src_dir, partname), 'rb') as infile:
                content += infile.read()
        
        print(f"Checking checksum for {current_set['name']}...")
        crc_val = hex(zlib.crc32(content))
        if crc_val != hex(int(current_set['crc'], 16)):
            print(f"CRC does not match - found {crc_val}, expected {hex(int(current_set['crc'], 16))} - skipping set!")
        else:
            print(f"Creating file for {current_set['name']}...")
            with open(os.path.join(dest_dir, f"{current_set['safename']}.iso"), 'wb') as outfile:
                outfile.write(content)



import os
import click
import logging
import sys
from rich.console import Console
from rich.markdown import Markdown

@click.command()
@click.option('--task', 'task', help = 'name of the transform set to run', required=True)
def details(task):
    """Get details about a specific Steam extraction task"""

    logging.basicConfig(stream=sys.stderr, level=logging.INFO)
    dir_path = os.path.dirname(os.path.realpath(__file__))
    md_path = os.path.join(dir_path, "..", "..", "..", "lib", "transforms", f'{task}.md')
    if not os.path.exists(md_path):
        print(md_path)
        raise Exception("Task not found to describe.")
    else:
        with open(md_path, "r") as curr_file:
            file_content = curr_file.read()
            console = Console()
            md = Markdown(file_content)
            console.print(md)





# Extraction Script for Capcom Fighting Collection

# General Extraction Process
# - Extract ARC Archive
# - Pull bin/ROMNAME file
# - Split it into parts using offsets/length
#   - Header (60b)
#   - MainCPU (???k)
#   - ??? inv gfx (???k) 
#   - AudioCPU (???k)
#   - QSound (???k)
# - Process each part
#   - maincpu: geometry
#   - gfx: geometry, deinterleave, endian swap?
#   - audiocpu: geometry
#   - qsound: geometry + endian swap

import re
import traceback
import glob
import zipfile
import logging
import os
import io

from gex.lib.archive import arc, ibis
from gex.lib.utils import blob

title = "Capcom Fighting Collection"
description = "NYI"
in_dir_desc = "CFC base folder (Ex. C:\Program Files (x86)\Steam\steamapps\common\CAPCOM FIGHTING COLLECTION)"

pkg_name_map = {
    "game_00.arc": "vampj",
    "game_01.arc": "dstlku",
    "game_10.arc": "vhuntjr2",
    "game_11.arc": "nwarru",
    "game_20.arc": "vsavj",
    "game_21.arc": "vsavu",
    "game_30.arc": "vhunt2",
    "game_40.arc": "vsav2",
    "game_50.arc": "cybotsj",
    "game_51.arc": "cybotsu",
    "game_60.arc": "spf2xj",
    "game_61.arc": "spf2tu",
    "game_70.arc": "pfghtj",
    "game_71.arc": "sgemf",
    "game_80.arc": "hsf2j",
    "game_81.arc": "hsf2",
    "game_90.arc": "warzard",
    "game_91.arc": "redearth",
}

def write_temp_file(contents, path):
    with open(path, 'wb') as out_file:
        out_file.write(contents)

def find_files(base_path):
    arc_path = os.path.join(base_path, "nativeDX11x64", "arc", "pc") 
    candidate_files = glob.glob(arc_path +'/game_*.arc')
    archive_list = []
    for candidate in candidate_files:
        if re.search(r'game_\d\d.arc', candidate):
            archive_list.append(candidate)
    return archive_list


################################################################################
# START Darkstalkers/Vampire: The Night Warriors                               #
################################################################################
# game_00.arc: Vampire: The Night Warriors (JP)
# game_01.arc: Darkstalkers: The Night Warriors

#   0x0000000   0x0000040       IBIS Header
#   0x0000040   0x0400040       maincpu - OK!
#   0x0400040   0x0800040       ???
#   0x0800040   0x1C00040       gfx - OK!
#   0x1C00040   0x1C48040       audiocpu - OK!
#   0x1C48040   0x1C50040       ???
#   0x1C50040   0x2050040       qsound - OK!

vam_gfx_filenames = [
    'vam.13m',
    'vam.14m',
    'vam.15m',
    'vam.16m',
    'vam.17m',
    'vam.18m',
    'vam.19m',
    'vam.20m'
]

vam_audiocpu_filenames = [
    'vam.01',
    'vam.02'
]

vam_qsound_filenames = [
    'vam.11m',
    'vam.12m'
]

def handle_vampj(merged_contents): 
    out_files = []   
    func_map = {}

    maincpu_filenames = [   
        "vamj.03a",
        "vamj.04b",
        "vamj.05a",
        "vamj.06a",
        "vamj.07a",
        "vamj.08a",
        "vamj.09a",
        "vamj.10a"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x1400000, vam_gfx_filenames, split=[0x400000, 0x100000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x1C00040, vam_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x1C50040, 0x400000, vam_qsound_filenames)

    out_files.append({'filename': 'vampj.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files


def handle_dstlku(merged_contents): 
    out_files = []   
    func_map = {}

    maincpu_filenames = [   
        "vamu.03b",
        "vamu.04b",
        "vamu.05b",
        "vamu.06b",
        "vamu.07b",
        "vamu.08b",
        "vamu.09b",
        "vamu.10b"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x1400000, vam_gfx_filenames, split=[0x400000, 0x100000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x1C00040, vam_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x1C50040, 0x400000, vam_qsound_filenames)

    out_files.append({'filename': 'dstlku.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files

################################################################################
# END Darkstalkers/Vampire: The Night Warriors                                 #
################################################################################


################################################################################
# START Vampire Hunter/Night Warriors: Darkstalkers' Revenge                   #
################################################################################
# game_10.arc: Vampire: The Night Warriors (JP)
# game_11.arc: Darkstalkers: The Night Warriors

#   0x0000000   0x0000040       IBIS Header
#   0x0000040   0x0400040       maincpu
#   0x0400040   0x0800040       ???
#   0x0800040   0x2800040       gfx
#   0x2800040   0x2848040       audiocpu
#   0x2848040   0x2850040       ???
#   0x2850040   0x2C50040       qsound

vph_gfx_filenames = [
    'vph.13m',
    'vph.14m',
    'vph.15m',
    'vph.16m',
    'vph.17m',
    'vph.18m',
    'vph.19m',
    'vph.20m'
]

vph_audiocpu_filenames = [
    'vph.01',
    'vph.02'
]

vph_qsound_filenames = [
    'vph.11m',
    'vph.12m'
]

def handle_vhuntjr2(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [   
        "vphj.03b",
        "vphj.04a",
        "vphj.05a",
        "vphj.06a",
        "vphj.07a",
        "vphj.08a",
        "vphj.09a",
        "vphj.10a"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x2000000, vph_gfx_filenames, split=[0x400000, 0x400000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x2800040, vph_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x2850040, 0x400000, vph_qsound_filenames)

    out_files.append({'filename': 'vhuntjr2.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files


def handle_nwarru(merged_contents): 
    out_files = []

    maincpu_filenames = [   
        "vphu.03f",
        "vphu.04c",
        "vphu.05e",
        "vphu.06c",
        "vphu.07b",
        "vphu.08b",
        "vphu.09b",
        "vphu.10b"
    ]

    func_map = {}
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x2000000, vph_gfx_filenames, split=[0x400000, 0x400000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x2800040, vph_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x2850040, 0x400000, vph_qsound_filenames)

    out_files.append({'filename': 'nwarru.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files

################################################################################
# END Vampire Hunter/Night Warriors: Darkstalkers' Revenge                     #
################################################################################


################################################################################
# START Vampire Savior: The Lord of Vampire                                    #
################################################################################
# game_20.arc: Vampire Savior: The Lord of Vampire (JP)
# game_21.arc: Darkstalkers: The Night Warriors

#   0x0000000   0x0000040       IBIS Header
#   0x0000040   0x0400040       maincpu
#   0x0400040   0x0800040       ???
#   0x0800040   0x2800040       gfx
#   0x2800040   0x2848040       audiocpu
#   0x2848040   0x2850040       ???
#   0x2850040   0x3050040       qsound

vm3_gfx_filenames = [
    'vm3.13m',
    'vm3.14m',
    'vm3.15m',
    'vm3.16m',
    'vm3.17m',
    'vm3.18m',
    'vm3.19m',
    'vm3.20m'
]

vm3_audiocpu_filenames = [
    'vm3.01',
    'vm3.02'
]

vm3_qsound_filenames = [
    'vm3.11m',
    'vm3.12m'
]

def handle_vsavj(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [   
        "vm3j.03d",
        "vm3j.04d",
        "vm3j.05a",
        "vm3j.06b",
        "vm3j.07b",
        "vm3j.08a",
        "vm3j.09b",
        "vm3j.10b"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x2000000, vm3_gfx_filenames, split=[0x400000, 0x400000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x2800040, vm3_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x2850040, 0x800000, vm3_qsound_filenames)

    out_files.append({'filename': 'vsavj.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files

def handle_vsavu(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [   
        "vm3u.03d",
        "vm3u.04d",
        "vm3.05a",
        "vm3.06a",
        "vm3.07b",
        "vm3.08a",
        "vm3.09b",
        "vm3.10b"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x2000000, vm3_gfx_filenames, split=[0x400000, 0x400000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x2800040, vm3_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x2850040, 0x800000, vm3_qsound_filenames)

    out_files.append({'filename': 'vsavu.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files

################################################################################
# END Vampire Savior: The Lord of Vampire                                      #
################################################################################


################################################################################
# START Vampire Hunter 2: Darkstalkers Revenge                                 #
################################################################################
# game_30.arc: Vampire Hunter 2: Darkstalkers Revenge (JP)

#   0x0000000   0x0000040       IBIS Header
#   0x0000040   0x0400040       maincpu
#   0x0400040   0x0800040       ???
#   0x0800040   0x2800040       gfx
#   0x2800040   0x2848040       audiocpu
#   0x2848040   0x2850040       ???
#   0x2850040   0x3050040       qsound

vh2_gfx_filenames = [
    'vh2.13m',
    'vh2.14m',
    'vh2.15m',
    'vh2.16m',
    'vh2.17m',
    'vh2.18m',
    'vh2.19m',
    'vh2.20m'
]

vh2_audiocpu_filenames = [
    'vh2.01',
    'vh2.02'
]

vh2_qsound_filenames = [
    'vh2.11m',
    'vh2.12m'
]

def handle_vhunt2(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [   
        "vh2j.03a",
        "vh2j.04a",
        "vh2j.05",
        "vh2j.06",
        "vh2j.07",
        "vh2j.08",
        "vh2j.09",
        "vh2j.10"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x2000000, vh2_gfx_filenames, split=[0x400000, 0x400000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x2800040, vh2_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x2850040, 0x800000, vh2_qsound_filenames)

    out_files.append({'filename': 'vhunt2.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files


################################################################################
# END Vampire Hunter 2: Darkstalkers Revenge                                   #
################################################################################


################################################################################
# START Vampire Savior 2: The Lord of Vampire                                  #
################################################################################
# game_40.arc: Vampire Savior 2: The Lord of Vampire (JP)

#   0x0000000   0x0000040       IBIS Header
#   0x0000040   0x0400040       maincpu
#   0x0400040   0x0800040       ???
#   0x0800040   0x2800040       gfx
#   0x2800040   0x2848040       audiocpu
#   0x2848040   0x2850040       ???
#   0x2850040   0x3050040       qsound

vs2_gfx_filenames = [
    'vs2.13m',
    'vs2.14m',
    'vs2.15m',
    'vs2.16m',
    'vs2.17m',
    'vs2.18m',
    'vs2.19m',
    'vs2.20m'
]

vs2_audiocpu_filenames = [
    'vs2.01',
    'vs2.02'
]

vs2_qsound_filenames = [
    'vs2.11m',
    'vs2.12m'
]

def handle_vsav2(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [   
        "vs2j.03",
        "vs2j.04",
        "vs2j.05",
        "vs2j.06",
        "vs2j.07",
        "vs2j.08",
        "vs2j.09",
        "vs2j.10"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x2000000, vs2_gfx_filenames, split=[0x400000, 0x400000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x2800040, vs2_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x2850040, 0x800000, vs2_qsound_filenames)

    out_files.append({'filename': 'vsav2.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files


################################################################################
# END Vampire Savior 2: The Lord of Vampire                                    #
################################################################################


################################################################################
# START Cyberbots: Fullmetal Madness                                           #
################################################################################
# game_50.arc: Cyberbots: Fullmetal Madness (JP)
# game_51.arc: Cyberbots: Fullmetal Madness

#   0x0000000   0x0000040       IBIS Header
#   0x0000040   0x0100040       maincpu
#   0x0400040   0x0800040       ???
#   0x0800040   0x2800040       gfx
#   0x2800040   0x2848040       audiocpu
#   0x2848040   0x2850040       ???
#   0x2850040   0x2C50040       qsound

cybots_gfx_filenames = [
    'cyb.13m',
    'cyb.14m',
    'cyb.15m',
    'cyb.16m',
    'cyb.17m',
    'cyb.18m',
    'cyb.19m',
    'cyb.20m'
]

cybots_audiocpu_filenames = [
    'cyb.01',
    'cyb.02'
]

cybots_qsound_filenames = [
    'cyb.11m',
    'cyb.12m'
]

def handle_cybotsj(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [   
        "cybj.03",
        "cybj.04",
        "cyb.05",
        "cyb.06",
        "cyb.07",
        "cyb.08",
        "cyb.09",
        "cyb.10"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x2000000, cybots_gfx_filenames, split=[0x400000, 0x400000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x2800040, cybots_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x2850040, 0x400000, cybots_qsound_filenames)

    out_files.append({'filename': 'cybotsj.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files

def handle_cybotsu(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [   
        "cybu.03",
        "cybu.04",
        "cyb.05",
        "cyb.06",
        "cyb.07",
        "cyb.08",
        "cyb.09",
        "cyb.10"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x2000000, cybots_gfx_filenames, split=[0x400000, 0x400000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x2800040, cybots_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x2850040, 0x400000, cybots_qsound_filenames)

    out_files.append({'filename': 'cybotsu.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files

################################################################################
# END Cyberbots: Fullmetal Madness                                             #
################################################################################


################################################################################
# START Super Puzzle Fighter II Turbo                                          #
################################################################################
# game_60.arc: Super Puzzle Fighter II X (JP)
# game_61.arc: Super Puzzle Fighter II Turbo

#   0x0000000   0x0000040       IBIS Header
#   0x0000040   0x0100040       maincpu
#   0x0100040   0x1000040       ???
#   0x1000040   0x1400040       gfx
#   0x1400040   0x1448040       audiocpu
#   0x1450040   0x1850040       qsound

pzf_gfx_filenames = [
    'pzf.14m',
    'pzf.16m',
    'pzf.18m',
    'pzf.20m'
]

pzf_audiocpu_filenames = [
    'pzf.01',
    'pzf.02'
]

pzf_qsound_filenames = [
    'pzf.11m',
    'pzf.12m'
]

# def handle_spf2xj(merged_contents): 
#     out_files = []
#     func_map = {}

#     maincpu_filenames = [   
#         "pzfj.03a",
#         "pzf.04"
#     ]
#     func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x100000, 2, maincpu_filenames)
#     func_map['gfx'] = ibis.gfx_cps2(0x1000040, 0x400000, pzf_gfx_filenames, split=[0x100000])
#     func_map['audiocpu'] = ibis.audiocpu_cps2(0x1400040, pzf_audiocpu_filenames)
#     func_map['qsound'] = ibis.qsound_cps2(0x1450040, 0x400000, pzf_qsound_filenames)

#     out_files.append({'filename': 'spf2xj.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

#     return out_files

# def handle_spf2tu(merged_contents): 
#     out_files = []
#     func_map = {}

#     maincpu_filenames = [   
#         "pzfu.03a",
#         "pzf.04"
#     ]
#     func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x100000, 2, maincpu_filenames)
#     func_map['gfx'] = ibis.gfx_cps2(0x1000040, 0x400000, pzf_gfx_filenames, split=[0x100000])
#     func_map['audiocpu'] = ibis.audiocpu_cps2(0x1400040, pzf_audiocpu_filenames)
#     func_map['qsound'] = ibis.qsound_cps2(0x1450040, 0x400000, pzf_qsound_filenames)

#     out_files.append({'filename': 'spf2tu.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

#     return out_files

################################################################################
# END Super Puzzle Fighter II Turbo                                            #
################################################################################

################################################################################
# START Super Gem Fighter Mini Mix                                             #
################################################################################
# game_70.arc: Pocket Fighter (JP)
# game_71.arc: Super Gem Fighter Mini Mix

#   0x0000000   0x0000040       IBIS Header
#   0x0000040   0x0280040       maincpu
#   0x0280040   0x0800040       ???
#   0x0800040   0x1C00040       gfx
#   0x1C00040   0x1C48040       audiocpu
#   0x1C50040   0x2550040       qsound

pcf_gfx_filenames = [
    'pcf.13m',
    'pcf.14m',
    'pcf.15m',
    'pcf.16m',
    'pcf.17m',
    'pcf.18m',
    'pcf.19m',
    'pcf.20m'
]

pcf_audiocpu_filenames = [
    'pcf.01',
    'pcf.02'
]

pcf_qsound_filenames = [
    'pcf.11m',
    'pcf.12m'
]

# def handle_pfghtj(merged_contents): 
#     out_files = []
#     func_map = {}

#     maincpu_filenames = [   
#         "pcfj.03",
#         "pcf.04",
#         "pcf.05",
#         "pcf.06",
#         "pcf.07"
#     ]
#     func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x280000, 5, maincpu_filenames)
#     func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x1400000, pcf_gfx_filenames, split=[0x400000, 0x100000])
#     func_map['audiocpu'] = ibis.audiocpu_cps2(0x1C00040, pcf_audiocpu_filenames)
#     func_map['qsound'] = ibis.qsound_cps2(0x1C50040, 0x800000, pcf_qsound_filenames)

#     out_files.append({'filename': 'pfghtj.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

#     return out_files

# def handle_sgemf(merged_contents): 
#     out_files = []
#     func_map = {}

#     maincpu_filenames = [   
#         "pcfu.03a",
#         "pcf.04"
#     ]
#     func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x280000, 5, maincpu_filenames)
#     func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x1400000, pcf_gfx_filenames, split=[0x400000, 0x100000])
#     func_map['audiocpu'] = ibis.audiocpu_cps2(0x1C00040, pcf_audiocpu_filenames)
#     func_map['qsound'] = ibis.qsound_cps2(0x1C50040, 0x800000, pcf_qsound_filenames)

#     out_files.append({'filename': 'sgemf.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

#     return out_files

################################################################################
# END Super Gem Fighter Mini Mix                                               #
################################################################################


################################################################################
# START Hyper Street Fighter II: Anniversary Edition                           #
################################################################################
# game_80.arc: Hyper Street Fighter II: Anniversary Edition (JP)
# game_81.arc: Hyper Street Fighter II: Anniversary Edition

#   0x0000000   0x0000040       IBIS Header
#   0x0000040   0x0400040       maincpu
#   0x0400040   0x0800040       ???
#   0x0800040   0x2800040       gfx
#   0x2800040   0x2848040       audiocpu
#   0x2848040   0x2850040       ???
#   0x2850040   0x3050040       qsound

hs2_gfx_filenames = [
    'hs2.13m',
    'hs2.15m',
    'hs2.17m',
    'hs2.19m'
]

hs2_audiocpu_filenames = [
    'hs2.01',
    'hs2.02'
]

hs2_qsound_filenames = [
    'hs2.11m'
]

def handle_hsf2j(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [   
        "hs2j.03",
        "hs2j.04",
        "hs2j.05",
        "hs2j.06",
        "hs2j.07",
        "hs2j.08",
        "hs2j.09",
        "hs2j.10"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x2000000, hs2_gfx_filenames, split=[0x800000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x2800040, hs2_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x2850040, 0x800000, hs2_qsound_filenames, num_chunks=1)

    out_files.append({'filename': 'hsf2j.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files

def handle_hsf2(merged_contents): 
    out_files = []
    func_map = {}

    maincpu_filenames = [   
        "hs2u.03",
        "hs2u.04",
        "hs2.05",
        "hs2.06",
        "hs2.07",
        "hs2.08",
        "hs2.09",
        "hs2.10"
    ]
    func_map['maincpu'] = ibis.maincpu_cps2(0x40, 0x400000, 8, maincpu_filenames)
    func_map['gfx'] = ibis.gfx_cps2(0x0800040, 0x2000000, hs2_gfx_filenames, split=[0x800000])
    func_map['audiocpu'] = ibis.audiocpu_cps2(0x2800040, hs2_audiocpu_filenames)
    func_map['qsound'] = ibis.qsound_cps2(0x2850040, 0x800000, hs2_qsound_filenames, num_chunks=1)

    out_files.append({'filename': 'hsf2.zip', 'contents': merged_rom_handler(merged_contents, func_map)})

    return out_files

################################################################################
# END Hyper Street Fighter II: Anniversary Edition                             #
################################################################################

################################################################################
# START Red Earth                                                              #
################################################################################
# game_90.arc: Warzard (JP)
# game_91.arc: Red Earth

# This is CPS3. 
# The actual ROM is small - only 512KB - but there doesn't appear to be an exact match offhand.
# The majority of the data is in a CHD - but there doesn't appear to be an exact match offhand.

# Warzard has a 'JACK' header, but Red Earth has an 'IBIS' header. So they are likely a similar format.

# There are very 'word like' sections that cryptanalysis can be done with.
# EX: 0x1A490
#    orig: M:noadnouT:yuT:e
#  endian: :MondaonTuy:Tue:
# This section is pretty clearly days of week, followed by date format strings and months of year.
# It LOOKS like every 4 charcters are backwards 
#     sep: M:no | adno | uT:y | uT:e
#  4 char: on:M | onda | y:Tu | e:Tue
#  merged: on:Monday:Tue:Tue
# Doing this to the whole file gets us a lot of readable text (Days/Months/Error Strings                                                                                                                                                                                                                                           )... 
# But that readable text isn't in the romfile or compressed CHD.
# Did they pre-decompress the CHD for performance?

# There are CD001 headers - is this an ISO IMAGE?! https://en.wikipedia.org/wiki/ISO_9660#Specifications
# 2 Tracks:
#  0x1A1CF: hCD001� - type 68, ver 0 (invalid)
#  0x1D80C: hCD001� - type 68, ver 0 (invalid)
# There would be 32,768 bytes (0x8000) in a 'system area' before the first CD001 header
# This would mean the image starts at 0x1A1CF - 0x8000 = 0x121CF
# This is 0x0 - 0x121CF is NOT big enough to fit the 0x80000 ROM. Nor is 0x0 - 0x1A1CF. 
# More importantly - this isn't a valid ISO 9660
# The volume versions are invalid, there's not a primary or terminator...

# def handle_warzard(merged_contents): 
#     out_files = []
#     func_map = {}


#     def endian(contents):
#         contents = blob.byte_shuffle(contents, 4, [3, 2, 1, 0])
#         return { 'endian.bin': contents }

#     func_map['endian'] = endian
#     out_files.append({'filename': 'warzard.zip', 'contents': merged_rom_handler(merged_contents, func_map)})
#     return out_files

# def handle_redearth(merged_contents): 
#     out_files = []
#     func_map = {}


#     def endian(contents):
#         contents = blob.byte_shuffle(contents, 4, [3, 2, 1, 0])
#         return { 'endian.bin': contents }

#     func_map['endian'] = endian
#     out_files.append({'filename': 'redearth.zip', 'contents': merged_rom_handler(merged_contents, func_map)})
#     return out_files

################################################################################
# END Red Earth                                                                #
################################################################################


def merged_rom_handler(merged_contents, func_map):
    new_data = dict()
    for func in func_map.values():
        new_data.update(func(bytearray(merged_contents)))

    # Build the new zip file
    new_contents = io.BytesIO()
    with zipfile.ZipFile(new_contents, "w") as new_archive:
        for name, data in new_data.items():
            new_archive.writestr(name, data)
    return new_contents.getvalue()


def main(game_base_dir, out_path):
    pak_files = find_files(game_base_dir)
    for file_path in pak_files:
        file_name = os.path.basename(file_path)
        pkg_name = pkg_name_map.get(file_name)
        if not pkg_name == None:
            logging.info(f"Extracting {file_name}: {pkg_name}") 
            try:
                with open(file_path, "rb") as curr_file:
                    file_content = bytearray(curr_file.read())
                    arc_contents = arc.extract(file_content)
                    output_files = []

                    # Get the bin entry
                    merged_rom_contents = None
                    for key, arc_content in arc_contents.items():
                        if arc_content['path'].startswith('bin'):
                            merged_rom_contents = arc_content['contents']

                    handler_func = globals().get(f'handle_{pkg_name}')

                    if merged_rom_contents != None and handler_func != None:
                        output_files = handler_func(merged_rom_contents)
                            
                        for output_file in output_files:
                            with open(os.path.join(out_path, output_file['filename']), "wb") as out_file:
                                out_file.write(output_file['contents'])
                    elif merged_rom_contents == None:
                        print("Could not find merged rom data in arc.")
                    elif handler_func == None:
                        print("Could not find matching handler function.")
            except Exception as e:
                traceback.print_exc()
                logging.warning(f'Error while processing {file_path}!') 
        else:
            logging.info(f'Skipping unmatched file {file_path}!') 
    logging.info("""
        Processing complete. 
    """)



import click

from .details import details
from .extract import extract
from .list import list

@click.group()
def steam():
    """Extract from Steam Apps"""
    pass

steam.add_command(details)
steam.add_command(extract)
steam.add_command(list)


import click
import logging
import importlib
import sys

@click.command()
@click.option('--srcdir', 'src_dir', help = 'path required by the transform set', required=True)
@click.option('--destdir', 'dest_dir', help = 'path to send reassembled ROMs to', required=True)
@click.option('--task', 'task', help = 'name of the transform set to run', required=True)
def extract(src_dir, dest_dir, task):
    """Run a task to pull from Steam app or Content pulls (via depot download)"""

    logging.basicConfig(stream=sys.stderr, level=logging.INFO)
    transform_module = importlib.import_module(f'gex.lib.transforms.{task}')
    transform_module.main(src_dir, dest_dir)



import magic

def enhanced_magic_from_path(in_file):
    magic_id = None
    try:
        magic_id = magic.from_file(in_file)
    except Exception as e:
        print(repr(e))

    if magic_id != 'data' and magic_id != None:
        return magic_id
    
    # Try to identify based on known file types magic bytes
    with open(in_file, 'rb') as file:
        content_start = file.read(2048)
        return enhanced_look(content_start)

def enhanced_magic_from_buffer(content_peek):
    magic_id = None
    try:
        magic_id = magic.from_buffer(content_peek)
    except Exception as e:
        print(repr(e))

    if magic_id != 'data' and magic_id != None:
        return magic_id
    
    # Try to identify based on known file types magic bytes
    return enhanced_look(content_peek)

def enhanced_look(content_peek):
    if content_peek[0:4] == b"KPKA":
        return "KPKA Archive, Capcom RE Engine"
    if content_peek[0:4] == b"IBIS":
        return "IBIS Archive, Capcom ROM Releases"
    if content_peek[0:3] == b"ARC":
        return "ARC Archive, Capcom MT Engine"
        
    return 'data'
            


import click

from .deinterleave import deinterleave
from .identify import identify
from .slice import slice

@click.group()
def file():
    """Generic file operations"""
    pass

file.add_command(deinterleave)
file.add_command(identify)
file.add_command(slice)